@@ -1,3 +1,25 @@
+// Copyright (C) 2004-2023 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
 #include "mupdf/fitz.h"
 
 #include "z-imp.h"
@@ -41,6 +63,7 @@ fz_save_pixmap_as_png(fz_context *ctx, fz_pixmap *pixmap, const char *filename)
 		writer = fz_new_png_band_writer(ctx, out);
 		fz_write_header(ctx, writer, pixmap->w, pixmap->h, pixmap->n, pixmap->alpha, pixmap->xres, pixmap->yres, 0, pixmap->colorspace, pixmap->seps);
 		fz_write_band(ctx, writer, pixmap->stride, pixmap->h, pixmap->samples);
+		fz_close_band_writer(ctx, writer);
 		fz_close_output(ctx, out);
 	}
 	fz_always(ctx)
@@ -68,6 +91,7 @@ fz_write_pixmap_as_png(fz_context *ctx, fz_output *out, const fz_pixmap *pixmap)
 	{
 		fz_write_header(ctx, writer, pixmap->w, pixmap->h, pixmap->n, pixmap->alpha, pixmap->xres, pixmap->yres, 0, pixmap->colorspace, pixmap->seps);
 		fz_write_band(ctx, writer, pixmap->stride, pixmap->h, pixmap->samples);
+		fz_close_band_writer(ctx, writer);
 	}
 	fz_always(ctx)
 	{
@@ -84,7 +108,7 @@ typedef struct png_band_writer_s
 	fz_band_writer super;
 	unsigned char *udata;
 	unsigned char *cdata;
-	uLong usize, csize;
+	size_t usize, csize;
 	z_stream stream;
 	int stream_ended;
 } png_band_writer;
@@ -147,12 +171,17 @@ png_write_header(fz_context *ctx, fz_band_writer *writer_, fz_colorspace *cs)
 
 	if (writer->super.s != 0)
 		fz_throw(ctx, FZ_ERROR_GENERIC, "PNGs cannot contain spot colors");
+	if (fz_colorspace_type(ctx, cs) == FZ_COLORSPACE_BGR)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "pixmap can not be bgr");
+	if (cs && !fz_colorspace_is_gray(ctx, cs) && !fz_colorspace_is_rgb(ctx, cs))
+		fz_throw(ctx, FZ_ERROR_GENERIC, "pixmap must be grayscale or rgb to write as png");
 
 	/* Treat alpha only as greyscale */
 	if (n == 1 && alpha)
 		alpha = 0;
+	n -= alpha;
 
-	switch (n - alpha)
+	switch (n)
 	{
 	case 1: color = (alpha ? 4 : 0); break; /* 0 = Greyscale, 4 = Greyscale + Alpha */
 	case 3: color = (alpha ? 6 : 2); break; /* 2 = RGB, 6 = RGBA */
@@ -187,6 +216,7 @@ png_write_band(fz_context *ctx, fz_band_writer *writer_, int stride, int band_st
 	unsigned char *dp;
 	int y, x, k, err, finalband;
 	int w, h, n;
+	size_t remain;
 
 	if (!out)
 		return;
@@ -201,21 +231,29 @@ png_write_band(fz_context *ctx, fz_band_writer *writer_, int stride, int band_st
 
 	if (writer->udata == NULL)
 	{
-		writer->usize = ((uLong)w * n + 1) * band_height;
-		/* Sadly the bound returned by compressBound is just for a
-		 * single usize chunk; if you compress a sequence of them
-		 * the buffering can result in you suddenly getting a block
-		 * larger than compressBound outputted in one go, even if you
-		 * take all the data out each time. */
-		writer->csize = compressBound(writer->usize);
-		writer->udata = Memento_label(fz_malloc(ctx, writer->usize), "png_write_udata");
-		writer->cdata = Memento_label(fz_malloc(ctx, writer->csize), "png_write_cdata");
+		size_t usize = w;
+
+		if (usize > SIZE_MAX / n - 1)
+			fz_throw(ctx, FZ_ERROR_GENERIC, "png data too large.");
+		usize = usize * n + 1;
+		if (usize > SIZE_MAX / band_height)
+			fz_throw(ctx, FZ_ERROR_GENERIC, "png data too large.");
+		usize *= band_height;
 		writer->stream.opaque = ctx;
 		writer->stream.zalloc = fz_zlib_alloc;
 		writer->stream.zfree = fz_zlib_free;
 		err = deflateInit(&writer->stream, Z_DEFAULT_COMPRESSION);
 		if (err != Z_OK)
 			fz_throw(ctx, FZ_ERROR_GENERIC, "compression error %d", err);
+		writer->usize = usize;
+		/* Now figure out how large a buffer we need to compress into.
+		 * deflateBound always expands a bit, and it's limited by being
+		 * a uLong rather than a size_t. */
+		writer->csize = writer->usize >= UINT32_MAX ? UINT32_MAX : deflateBound(&writer->stream, (uLong)writer->usize);
+		if (writer->csize < writer->usize || writer->csize > UINT32_MAX) /* Check for overflow */
+			writer->csize = UINT32_MAX;
+		writer->udata = Memento_label(fz_malloc(ctx, writer->usize), "png_write_udata");
+		writer->cdata = Memento_label(fz_malloc(ctx, writer->csize), "png_write_cdata");
 	}
 
 	dp = writer->udata;
@@ -225,24 +263,13 @@ png_write_band(fz_context *ctx, fz_band_writer *writer_, int stride, int band_st
 		/* Unpremultiply data */
 		for (y = 0; y < band_height; y++)
 		{
-			int prev[FZ_MAX_COLORS];
-			*dp++ = 1; /* sub prediction filter */
+			*dp++ = 0; /* none prediction filter */
 			for (x = 0; x < w; x++)
 			{
 				int a = sp[n-1];
 				int inva = a ? 256*255/a : 0;
-				int p;
 				for (k = 0; k < n-1; k++)
-				{
-					int v = (sp[k] * inva + 128)>>8;
-					p = x ? prev[k] : 0;
-					prev[k] = v;
-					v -= p;
-					dp[k] = v;
-				}
-				p = x ? prev[k] : 0;
-				prev[k] = a;
-				a -= p;
+					dp[k] = (sp[k] * inva + 128)>>8;
 				dp[k] = a;
 				sp += n;
 				dp += n;
@@ -254,16 +281,11 @@ png_write_band(fz_context *ctx, fz_band_writer *writer_, int stride, int band_st
 	{
 		for (y = 0; y < band_height; y++)
 		{
-			*dp++ = 1; /* sub prediction filter */
+			*dp++ = 0; /* none prediction filter */
 			for (x = 0; x < w; x++)
 			{
 				for (k = 0; k < n; k++)
-				{
-					if (x == 0)
-						dp[k] = sp[k];
-					else
-						dp[k] = sp[k] - sp[k-n];
-				}
+					dp[k] = sp[k];
 				sp += n;
 				dp += n;
 			}
@@ -271,38 +293,38 @@ png_write_band(fz_context *ctx, fz_band_writer *writer_, int stride, int band_st
 		}
 	}
 
-	writer->stream.next_in = (Bytef*)writer->udata;
-	writer->stream.avail_in = (uInt)(dp - writer->udata);
+	remain = dp - writer->udata;
+	dp = writer->udata;
+
 	do
 	{
+		size_t eaten;
+
+		writer->stream.next_in = dp;
+		writer->stream.avail_in = (uInt)(remain <= UINT32_MAX ? remain : UINT32_MAX);
 		writer->stream.next_out = writer->cdata;
-		writer->stream.avail_out = (uInt)writer->csize;
+		writer->stream.avail_out = writer->csize <= UINT32_MAX ? (uInt)writer->csize : UINT32_MAX;
 
-		if (!finalband)
-		{
-			err = deflate(&writer->stream, Z_NO_FLUSH);
-			if (err != Z_OK)
-				fz_throw(ctx, FZ_ERROR_GENERIC, "compression error %d", err);
-		}
-		else
-		{
-			err = deflate(&writer->stream, Z_FINISH);
-			if (err == Z_OK)
-			{
-				/* more output space needed, try again */
-				writer->cdata = Memento_label(fz_realloc(ctx, writer->cdata, writer->csize << 2), "realloc png_write_cdata");
-				writer->csize <<= 2;
-				continue;
-			}
+		err = deflate(&writer->stream, (finalband && remain == writer->stream.avail_in) ? Z_FINISH : Z_NO_FLUSH);
+		if (err != Z_OK && err != Z_STREAM_END)
+			fz_throw(ctx, FZ_ERROR_GENERIC, "compression error %d", err);
 
-			if (err != Z_STREAM_END)
-				fz_throw(ctx, FZ_ERROR_GENERIC, "compression error %d", err);
-		}
+		/* We are guaranteed that writer->stream.next_in will have been updated for the
+		 * data that has been eaten. */
+		eaten = (writer->stream.next_in - dp);
+		remain -= eaten;
+		dp += eaten;
 
+		/* We are guaranteed that writer->stream.next_out will have been updated for the
+		 * data that has been written. */
 		if (writer->stream.next_out != writer->cdata)
 			putchunk(ctx, out, "IDAT", writer->cdata, writer->stream.next_out - writer->cdata);
+
+		/* Zlib only guarantees to have finished when we have no more data to feed in, and
+		 * the last call to deflate did not return with avail_out == 0. (i.e. no more is
+		 * buffered internally.) */
 	}
-	while (writer->stream.avail_out == 0);
+	while (remain != 0 || writer->stream.avail_out == 0);
 }
 
 static void
