@@ -1,11 +1,17 @@
 #include "jsi.h"
-#include "jsparse.h"
-#include "jscompile.h"
-#include "jsvalue.h"
-#include "jsrun.h"
-#include "jsbuiltin.h"
 
 #include <assert.h>
+#include <errno.h>
+
+static int js_ptry(js_State *J) {
+	if (J->trytop == JS_TRYLIMIT) {
+		J->stack[J->top].t.type = JS_TLITSTR;
+		J->stack[J->top].u.litstr = "exception stack overflow";
+		++J->top;
+		return 1;
+	}
+	return 0;
+}
 
 static void *js_defaultalloc(void *actx, void *ptr, int size)
 {
@@ -30,6 +36,8 @@ static void js_defaultpanic(js_State *J)
 
 int js_ploadstring(js_State *J, const char *filename, const char *source)
 {
+	if (js_ptry(J))
+		return 1;
 	if (js_try(J))
 		return 1;
 	js_loadstring(J, filename, source);
@@ -39,6 +47,8 @@ int js_ploadstring(js_State *J, const char *filename, const char *source)
 
 int js_ploadfile(js_State *J, const char *filename)
 {
+	if (js_ptry(J))
+		return 1;
 	if (js_try(J))
 		return 1;
 	js_loadfile(J, filename);
@@ -49,6 +59,10 @@ int js_ploadfile(js_State *J, const char *filename)
 const char *js_trystring(js_State *J, int idx, const char *error)
 {
 	const char *s;
+	if (js_ptry(J)) {
+		js_pop(J, 1);
+		return error;
+	}
 	if (js_try(J)) {
 		js_pop(J, 1);
 		return error;
@@ -58,6 +72,54 @@ const char *js_trystring(js_State *J, int idx, const char *error)
 	return s;
 }
 
+double js_trynumber(js_State *J, int idx, double error)
+{
+	double v;
+	if (js_ptry(J)) {
+		js_pop(J, 1);
+		return error;
+	}
+	if (js_try(J)) {
+		js_pop(J, 1);
+		return error;
+	}
+	v = js_tonumber(J, idx);
+	js_endtry(J);
+	return v;
+}
+
+int js_tryinteger(js_State *J, int idx, int error)
+{
+	int v;
+	if (js_ptry(J)) {
+		js_pop(J, 1);
+		return error;
+	}
+	if (js_try(J)) {
+		js_pop(J, 1);
+		return error;
+	}
+	v = js_tointeger(J, idx);
+	js_endtry(J);
+	return v;
+}
+
+int js_tryboolean(js_State *J, int idx, int error)
+{
+	int v;
+	if (js_ptry(J)) {
+		js_pop(J, 1);
+		return error;
+	}
+	if (js_try(J)) {
+		js_pop(J, 1);
+		return error;
+	}
+	v = js_toboolean(J, idx);
+	js_endtry(J);
+	return v;
+}
+
 static void js_loadstringx(js_State *J, const char *filename, const char *source, int iseval)
 {
 	js_Ast *P;
@@ -69,7 +131,7 @@ static void js_loadstringx(js_State *J, const char *filename, const char *source
 	}
 
 	P = jsP_parse(J, filename, source);
-	F = jsC_compile(J, P);
+	F = jsC_compilescript(J, P, iseval ? J->strict : J->default_strict);
 	jsP_freeparse(J);
 	js_newscript(J, F, iseval ? (J->strict ? J->E : NULL) : J->GE);
 
@@ -89,41 +151,42 @@ void js_loadstring(js_State *J, const char *filename, const char *source)
 void js_loadfile(js_State *J, const char *filename)
 {
 	FILE *f;
-	char *s;
+	char *s, *p;
 	int n, t;
 
 	f = fopen(filename, "rb");
 	if (!f) {
-		js_error(J, "cannot open file: '%s'", filename);
+		js_error(J, "cannot open file '%s': %s", filename, strerror(errno));
 	}
 
 	if (fseek(f, 0, SEEK_END) < 0) {
 		fclose(f);
-		js_error(J, "cannot seek in file: '%s'", filename);
+		js_error(J, "cannot seek in file '%s': %s", filename, strerror(errno));
 	}
 
 	n = ftell(f);
 	if (n < 0) {
 		fclose(f);
-		js_error(J, "cannot tell in file: '%s'", filename);
+		js_error(J, "cannot tell in file '%s': %s", filename, strerror(errno));
 	}
 
 	if (fseek(f, 0, SEEK_SET) < 0) {
 		fclose(f);
-		js_error(J, "cannot seek in file: '%s'", filename);
+		js_error(J, "cannot seek in file '%s': %s", filename, strerror(errno));
 	}
 
-	s = js_malloc(J, n + 1); /* add space for string terminator */
-	if (!s) {
+	if (js_try(J)) {
 		fclose(f);
-		js_error(J, "cannot allocate storage for file contents: '%s'", filename);
+		js_throw(J);
 	}
+	s = js_malloc(J, n + 1); /* add space for string terminator */
+	js_endtry(J);
 
 	t = fread(s, 1, (size_t)n, f);
 	if (t != n) {
 		js_free(J, s);
 		fclose(f);
-		js_error(J, "cannot read data from file: '%s'", filename);
+		js_error(J, "cannot read data from file '%s': %s", filename, strerror(errno));
 	}
 
 	s[n] = 0; /* zero-terminate string containing file data */
@@ -134,7 +197,15 @@ void js_loadfile(js_State *J, const char *filename)
 		js_throw(J);
 	}
 
-	js_loadstring(J, filename, s);
+	/* skip first line if it starts with "#!" */
+	p = s;
+	if (p[0] == '#' && p[1] == '!') {
+		p += 2;
+		while (*p && *p != '\n')
+			++p;
+	}
+
+	js_loadstring(J, filename, p);
 
 	js_free(J, s);
 	fclose(f);
@@ -143,6 +214,11 @@ void js_loadfile(js_State *J, const char *filename)
 
 int js_dostring(js_State *J, const char *source)
 {
+	if (js_ptry(J)) {
+		js_report(J, "exception stack overflow");
+		js_pop(J, 1);
+		return 1;
+	}
 	if (js_try(J)) {
 		js_report(J, js_trystring(J, -1, "Error"));
 		js_pop(J, 1);
@@ -158,6 +234,11 @@ int js_dostring(js_State *J, const char *source)
 
 int js_dofile(js_State *J, const char *filename)
 {
+	if (js_ptry(J)) {
+		js_report(J, "exception stack overflow");
+		js_pop(J, 1);
+		return 1;
+	}
 	if (js_try(J)) {
 		js_report(J, js_trystring(J, -1, "Error"));
 		js_pop(J, 1);
@@ -204,7 +285,7 @@ js_State *js_newstate(js_Alloc alloc, void *actx, int flags)
 	js_State *J;
 
 	assert(sizeof(js_Value) == 16);
-	assert(soffsetof(js_Value, type) == 15);
+	assert(soffsetof(js_Value, t.type) == 15);
 
 	if (!alloc)
 		alloc = js_defaultalloc;
@@ -234,6 +315,12 @@ js_State *js_newstate(js_Alloc alloc, void *actx, int flags)
 
 	J->gcmark = 1;
 	J->nextref = 0;
+	J->gcthresh = 0; /* reaches stability within ~ 2-5 GC cycles */
+
+	if (js_try(J)) {
+		js_freestate(J);
+		return NULL;
+	}
 
 	J->R = jsV_newobject(J, JS_COBJECT, NULL);
 	J->G = jsV_newobject(J, JS_COBJECT, NULL);
@@ -242,5 +329,6 @@ js_State *js_newstate(js_Alloc alloc, void *actx, int flags)
 
 	jsB_init(J);
 
+	js_endtry(J);
 	return J;
 }
